Imports System.Math
Imports System.Runtime.InteropServices

Public Class Ran3
    Private Const MBIG As Integer = Integer.MaxValue
    Private Const MSEED As Integer = Integer.MaxValue \ 13
    Private Const NTAB As Integer = 55

    Private iseed As Integer
    Private inext, inextp As Integer
    Private ma(NTAB - 1) As Integer

    <StructLayout(LayoutKind.Explicit)> _
    Private Structure DU
        <FieldOffset(0)> Public D As Double
        <FieldOffset(0)> Public L As Long
        <FieldOffset(0)> Public I0 As Integer
        <FieldOffset(3)> Public I3 As Integer
        <FieldOffset(4)> Public I4 As Integer

        Public Const FracBitLength As Integer = 52I
        Public Const ImplicitBit As Long = 1L << FracBitLength
        Public Const ExplicitTopBit As Long = ImplicitBit \ 2
        Public Const MaskGetFrac As Long = ImplicitBit - 1
    End Structure

    Public Sub New()
        Initialize()
    End Sub
    Public Sub New(ByVal seed As Integer)
        Initialize(seed)
    End Sub

    Public Sub Initialize()
        Dim s1 As Integer = My.Computer.Clock.TickCount
        Dim s2 As Integer

        Dim c As Integer = 1
        Do
            s2 += (s1 And 1)
            If c = 31 Then Exit Do
            s1 = s1 >> 1
            s2 = s2 << 1
            c += 1
        Loop

        Initialize(s2)
    End Sub
    Public Sub Initialize(ByVal seed As Integer)
        iseed = seed

        Dim mj As Integer = Abs(MSEED - Abs(seed)) Mod MBIG
        ma(NTAB - 1) = mj

        Dim mk As Integer = 1

        For i As Integer = 0 To NTAB - 2
            Dim ii As Integer = ((21 * (i + 1)) Mod NTAB) - 1
            ma(ii) = mk
            mk = mj - mk
            If mk < 0 Then mk += MBIG
            mj = ma(ii)
        Next

        For k As Integer = 1 To 4
            For i As Integer = 0 To NTAB - 1
                ma(i) -= ma((i + 31) Mod 55)
                If ma(i) < 0 Then ma(i) += MBIG
            Next
        Next

        inext = -1
        inextp = 30
    End Sub

    Public Function NextRaw() As Integer
        inext = (inext + 1) Mod NTAB
        inextp = (inextp + 1) Mod NTAB

        Dim mj As Integer = ma(inext) - ma(inextp)
        If mj < 0 Then mj += MBIG
        ma(inext) = mj

        Return mj
    End Function
    Public Function NextDoubleRaw() As Double
        Dim r As DU
        r.L = CLng(NextRaw()) << 21
        r.I4 += &H3FF00000I
        Return r.D - 1.0
    End Function
    Public Sub NextBytesRaw(ByVal b() As Byte)
        If b Is Nothing Then Exit Sub
        For I As Integer = 0 To b.Length - 1
            b(I) = CByte(NextRaw() >> 23)
        Next
    End Sub

    Protected Overridable Function Sample() As Double
        Dim r As DU
        r.I0 = (NextRaw() >> 5)
        r.I3 += (NextRaw() And &H7FFFFFE0I) >> 3
        r.I4 += &H3FF00000I
        Return r.D - 1.0!
    End Function

    Public Function NextDouble() As Double
        Return Sample()
    End Function

    Public Function [Next]() As Integer
        Dim r As DU
        r.D = Sample() * Integer.MaxValue

        Dim sd As Integer = 1075 - (r.I4 >> 20)
        If sd > 52 Then
            Return 0
        Else
            Return CInt(((r.L And DU.MaskGetFrac) Or DU.ImplicitBit) >> sd)
        End If
    End Function
    Public Function [Next](ByVal maxValue As Integer) As Integer
        If maxValue < 0 Then _
            Throw New System.ArgumentOutOfRangeException _
                        ("maxValue", maxValue, "The argument must not be negative.")

        Dim r As DU
        r.D = Sample() * maxValue

        Dim sd As Integer = 1075 - (r.I4 >> 20)
        If sd > 52 Then
            Return 0
        Else
            Return CInt(((r.L And DU.MaskGetFrac) Or DU.ImplicitBit) >> sd)
        End If
    End Function
    Public Function [Next](ByVal minValue As Integer, ByVal maxValue As Integer) As Integer
        If maxValue < minValue Then _
            Throw New System.ArgumentOutOfRangeException _
                        ("minValue/maxValue", _
                         "The minValue must not be greater than the maxValue.")

        Dim r As DU
        r.D = Sample() * (CDbl(maxValue) - CDbl(minValue))

        Dim sd As Integer = 1075 - (r.I4 >> 20)
        If sd > 52 Then
            Return minValue
        Else
            Return CInt(minValue + (((r.L And DU.MaskGetFrac) Or DU.ImplicitBit) >> sd))
        End If
    End Function

    Public Sub NextBytes(ByVal b() As Byte)
        If b Is Nothing Then Exit Sub
        For I As Integer = 0 To b.Length - 1
            b(I) = CByte([Next]() >> 23)
        Next
    End Sub

    Public ReadOnly Property Seed() As Integer
        Get
            Return iseed
        End Get
    End Property
End Class

